 /* 
  * UAE - The Un*x Amiga Emulator
  * 
  * Debugger
  * 
  * (c) 1995 Bernd Schmidt
  *
  */

#include "sysconfi.h"
#include "sysdeps.h"

#include <ctype.h>
#include <signal.h>
#include <conio.h>
#include <dpmi.h>
#include <go32.h>
#include <dir.h>
#include <process.h>

#include "config.h"
#include "options.h"
#include "debug.h"
#include "memory2.h"
#include "Memory.h"
#include "custom.h"
#include "readcpu.h"
#include "newcpu.h"
#include "r_newcpu.h"
#include "uaepatch.h"
#include "ram.h"
#include "prefs.h"

#include "sound.h"
#include "IWM.h"
#include "RTC.h"
#include "SCC.h"
#include "VIA.h"

#include "IBM-SND.h"
#include "IBM-TIME.h"
#include "IBM-VGA.h"


#include "unix/universa.h"

extern int quit_program;
extern int use_debugger;
extern void MacInterrupt (void);

FILE *inStream = stdin;
FILE *ouStream = stdout;

static   CPTR skipaddr;
static   int  do_skip;
volatile CPTR breakaddr=0xffffffff;

int debugger_active = 0;
int _debugging = 0;
int Insert_Disk=0;
int RESET_SW = 0;
int GUI=0;
int INTERRUPT = 0;
int KeyLed=0;
int Dump=0;
int go=0;
int iLeds;

// BMP header information
UBYTE BMPhdr[] = {
   0x42, 0x4D,                // Signature = "BM"
   0x40, 0x96, 0x00, 0x00,    // File size = 640*480/8+64
   0x00, 0x00,                // r1        = 0
   0x00, 0x00,                // r2        = 0
   0x40, 0x00, 0x00, 0x00,    // Offset    = 64
   0x28, 0x00, 0x00, 0x00,    // Hsize     = 40
   0x80, 0x02, 0x00, 0x00,    // Width     = 640
   0xE0, 0x01, 0x00, 0x00,    // Height    = 480
   0x01, 0x00,                // Planes    = 1  
   0x01, 0x00,                // BPP       = 1
   0x00, 0x00, 0x00, 0x00,    // Compress  = 0
   0x00, 0x00, 0x00, 0x00,    // ISize     = 0
   0x00, 0x00, 0x00, 0x00,    // X/M       = 0
   0x00, 0x00, 0x00, 0x00,    // Y/M       = 0
   0x00, 0x00, 0x00, 0x00,    // Colours   = 0
   0x00, 0x00, 0x00, 0x00,    // Imp. Clrs = 0
   0x00, 0x00, 0x00,          // Color 0   = Black
   0xFF, 0xFF, 0xFF,          // Color 1   = White
   0xFF, 0xFF, 0xFF,          // Color 2   = White
   0x00                       // End of Header
};


// extern declarations of hardware variables
extern struct IWM_Ty IWM;
extern UBYTE Keyboard_Data;
extern UBYTE Keyboard_Enabled;
extern struct RTC_Ty RTC;
extern struct SCC_Ty SCC;
extern UBYTE SCC_Mode;
extern UBYTE SCC_Reg;
extern UBYTE Sony_Installed;
extern UBYTE Sony_Opened;
extern CPTR Sony_Table[5];
extern CPTR ParamBlk;
extern CPTR DeviceCtl;
extern int Floppy;
extern int Installed;
extern struct Sound_Ty Sound;
extern struct VIA_Ty VIA;
extern UBYTE T1_Active;
extern UBYTE T2_Active;
extern UBYTE SR_Active;
extern UBYTE SR_Timer_Active;
extern ULONG SR_Time;


// prototypes
void InsertDisk (void);
void DumpScreen (void);

static RETSIGTYPE sigbrkhandler(int foo)
{
   extern void EnterDebug(void);

   if (thePrefs.Sound) {
      Sound_Pause();
   }

   Timer_Uninstall();

   if (!GUI) {
      VGA_EnterDebug();
   }

   do_skip = 0;
   debugger_active = 1;
   _debugging = 1;
 
   broken_in = 1;   
   regs.spcflags |= SPCFLAG_BRK;
   signal(SIGINT, sigbrkhandler);
}

static void ignore_ws(char **c)
{
   while (**c && isspace(**c))
      (*c)++;
}

static ULONG readhex(char **c)
{
   ULONG val = 0;
   char nc;

   ignore_ws(c);

   while (isxdigit(nc = **c)) {
      (*c)++;
      val *= 16;
      nc = toupper(nc);
      if (isdigit(nc))
         val += nc - '0';
      else
          val += nc - 'A' + 10;
   }

   return val;
}

static char next_char(char **c)
{
   ignore_ws(c);
   return *(*c)++;
}

static int more_params(char **c)
{
   ignore_ws(c);
   return (**c) != 0;
}

static void dumpmem(CPTR addr, CPTR *nxmem, int lines)
{
   broken_in = 0;
   for (;lines-- && !broken_in;) {
      int i, j;

      if (ouStream==stdout) {
         cprintf("%08lx    ", addr);

         for (i=0; i<8; i++) {
            cprintf("%04x ", get_word(addr));
            j=wherex();
            gotoxy(56+(i*2), wherey());
            if (get_byte(addr)==0 || get_byte(addr)==7 || get_byte(addr)==8 ||
                get_byte(addr)==9 || get_byte(addr)==10|| get_byte(addr)==13||
                get_byte(addr)==32|| get_byte(addr)==255)
               cprintf(".");
            else
               cprintf("%c", (char) get_byte(addr));

            if (get_byte(addr+1)==0 || get_byte(addr+1)==7 || get_byte(addr+1)==8 ||
                get_byte(addr+1)==9 || get_byte(addr+1)==10|| get_byte(addr+1)==13||
                get_byte(addr+1)==32|| get_byte(addr+1)==255)
               cprintf(".");
            else
               cprintf("%c", (char) get_byte(addr+1));

            gotoxy(j, wherey());
            addr+=2;
         }
         printf("\n");
      } else {
         fprintf(ouStream, "%08lx ", addr);
         for (i=0; i< 8; i++) {
            fprintf(ouStream, "%04x ", get_word(addr));

            addr += 2;
         }
         fprintf(ouStream, "\n");
      }
   }
   *nxmem = addr;
}


void InsertDisk(void)
{
   UWORD  drivenum=Insert_Disk;
   char  *diskfile = malloc(255);
   int    ret=0;

   extern int  Sony_Disk_Inserted (UWORD Drive_No);
   extern void Sony_Insert (UWORD Drive_No, char *filename);

   if (drivenum==1) {
      Sony_Insert(0x0001, "A:");
   } else {
      fprintf(stderr, "Mount disk-image in drive %d : ", drivenum-1);
      fflush(stderr);
      fscanf(stdin, "%s", diskfile);
      fflush(stdin);

      if (!Sony_Disk_Inserted(drivenum)) {
         Sony_Insert(drivenum, (char *)diskfile);

         if (Sony_Disk_Inserted(drivenum)) {
            strcpy(thePrefs.DrivePath[drivenum-1], diskfile);
         }
      } else {
         fprintf(stderr, "There is already a disk mounted in drive %d.\n", drivenum-1);
      }
   }

   free(diskfile);

   return;
}

void DumpScreen (void)
{
   struct ffblk f;
   FILE *filBMP;
   unsigned VidRAM = 0xa0000;
   short int x, y;
   char  a;
   _go32_dpmi_registers Regs;
   char  filename[12];
   short int num=-1;
   int   done=0;

   strcpy(filename, "vMac0000.bmp");

   while (!done) {
     num++;
     filename[7]=48+(char) (num % 10);
     filename[6]=48+(char) ((num % 100) / 10);
     filename[5]=48+(char) ((num % 1000) / 100);
     filename[4]=48+(char) (num / 1000);

     done = findfirst(filename, &f, FA_ARCH|FA_RDONLY);
   }

   filBMP = fopen(filename, "wb+");
   fseek(filBMP, 0L, SEEK_SET);
   fwrite(&BMPhdr, sizeof(BMPhdr), 1, filBMP);

   for (y=479; y>=0; y--) {
      for (x=0; x<640; x+=8) {
         dosmemget((unsigned) (VidRAM+((640*y+x)/8)), 1, &a);
         fwrite(&a, 1, 1, filBMP);
      }
   }

   fclose(filBMP);
}

void debug(void)
{
   char input[80];
   CPTR nextpc,nxdis,nxmem;
   int  i;

   signal(SIGINT, sigbrkhandler);

   if (!use_debugger) {
      go=1;
   } else {
      raise(SIGINT);

      fprintf(ouStream, "debugging...\n");
      do_skip = 0;

      m68k_dumpstate(&nextpc);
      nxdis = nextpc; nxmem = 0;
   }

   for (;;) {
      char cmd,*inptr;

      if (!go) {
         if (inStream==stdin)
            fprintf(stderr, ">");

         fgets(input, 80, inStream);

         inptr = input;
         cmd = next_char(&inptr);
      } else {
         cmd = 'g';
      }

      switch(cmd) {
         case 'm': {
            ULONG maddr;
            int lines;

            if (more_params(&inptr))
               maddr = readhex(&inptr); 
            else 
               maddr = nxmem;

            if (more_params(&inptr))
               lines = readhex(&inptr); 
            else 
               lines = 16;

            dumpmem(maddr, &nxmem, lines);

            break;
         }

         case 'd': {
            ULONG daddr;
            int count;
		
            if (more_params(&inptr))
               daddr = readhex(&inptr);
            else 
               daddr = nxdis;

            if (more_params(&inptr))
               count = readhex(&inptr); 
            else
               count = 10;

            m68k_disasm(daddr, &nxdis, count);
            break;
         }

         case 'w': {
            ULONG maddr;
            int value;
            CPTR PC=m68k_getpc();

            if (more_params(&inptr))
               maddr = readhex(&inptr); 
            else 
               break;

            if (more_params(&inptr))
               value = readhex(&inptr); 
            else
               break;

            ROMmem_write_enable_set(on);
            put_byte(maddr, value);
            ROMmem_write_enable_set(off);

            m68k_setpc(PC);
            break;
         }

         case 'W': {
            UBYTE reg;
            ULONG value;
            CPTR PC=m68k_getpc();

            if (more_params(&inptr))
               reg = readhex(&inptr); 
            else 
               break;

            if (more_params(&inptr))
               value = readhex(&inptr); 
            else
               break;

            if ((((reg & 0xF0)!=0xA0) && ((reg & 0xF0)!=0xD0)) || ((reg & 0x0F)>7)) {
               fprintf(ouStream, "Invalid register.\n");
            } else {
               if ((reg & 0xF0)==0xA0)
                  m68k_areg(regs, reg & 0x0F) = value;
               else
                  m68k_dreg(regs, reg & 0x0F) = value;                 
            }

            m68k_setpc(PC);
            break;
         }

         case 'i': {
            extern int ROMmem_write_enable;
            if (ROMmem_write_enable) {
               ROMmem_write_enable_set(off);
               fprintf(ouStream, "ROM write access disabled.\n");
            } else {
               ROMmem_write_enable_set(on);
               fprintf(ouStream, "ROM write access enabled.\n");
            }
         }

         case 'r': {
            m68k_dumpstate(&nextpc);
            break;
         }

         case 'g': {
            if (use_debugger) {
               if (more_params (&inptr))
                  m68k_setpc (readhex (&inptr));
            }

            debugger_active = 0;
            _debugging = 0;

            go=0;

            Timer_Install();

            if (!GUI) {
               VGA_InstallGraphics();
            }

            GUI=0;

            regs.spcflags &= ~SPCFLAG_BRK;
            quit_program=0;

            if (ouStream!=stdout) {
               fclose(ouStream);
               ouStream=stdout;
            }

            if (inStream!=stdin) {
               fclose(inStream);
               inStream=stdin;
            }

            m68k_go(1);
            signal(SIGINT, sigbrkhandler);
            raise(SIGINT);

            if (Insert_Disk>0) {
               InsertDisk();
               Insert_Disk=0;
               go=1;
               use_debugger=0;
            } else {
               if (RESET_SW) {
                  extern void Reset_Emulator(void);
          
                  Reset_Emulator();
                  RESET_SW=0;
                  go=1;
                  use_debugger=0;
               } else {
                  if (Dump) {
                     DumpScreen();
                     Dump=0;
                     go=1;
                     use_debugger=0;
                  } else {
                     if (INTERRUPT) {
                        MacInterrupt();
                        INTERRUPT=0;
                        go=1;
                        use_debugger=0;
                     } else {
                        fprintf(ouStream, "debugging...\n");
                        m68k_dumpstate(&nextpc);
                        nxdis = nextpc; nxmem = 0;
                     }
                  }
               }
            }
            break;
         }

         case 't': {
            reentrant_m68k_step();
            m68k_dumpstate(&nextpc);
            break;
         }

         case 'z': {
            CPTR stop=m68k_getpc();

            if (more_params (&inptr)) {
               stop = readhex(&inptr);

               do {
                  reentrant_m68k_step();
               } while (stop!=m68k_getpc());
            }
            break;
         }

         case 'Z': {
            int cnt, i;

            if (more_params (&inptr)) {
               cnt = readhex(&inptr);

               for (i=0; i<cnt; i++)
                  reentrant_m68k_step();
            }
            break;
         }

         case 'f': {
            CPTR stop;

            if (more_params (&inptr)) {
               stop = readhex(&inptr);

               while (stop!=m68k_getpc()) {
                  reentrant_m68k_step();
               }
            }
            break;
         }

         case 'b': {
            CPTR stop;

            if (more_params (&inptr)) {
               stop = readhex(&inptr);

               if (breakaddr==stop) {
                  breakaddr=0xffffffff;
                  fprintf(ouStream, "Break-point removed.\n");
               } else {
                  breakaddr=stop;
                  fprintf(ouStream, "Break-point set at %08x.\n", breakaddr);
               }
            }
            break;
         }

         case '1': {
            FILE *RAMFile;
            char *MacRAM = (char *) get_real_address (kRAM_Size-kRAM_Size);
            CPTR i;
            ULONG x;

            RAMFile = fopen("vMac.RAM", "wb+");

            // Dump CPU state
            fwrite(&regs, sizeof(regs), 1, RAMFile);
            fwrite(&lastint_regs, sizeof(lastint_regs), 1, RAMFile);

            // Dump Hardware chips information and misc. control variables
            fwrite(&IWM, sizeof(IWM_Ty), 1, RAMFile);
            fwrite(&Keyboard_Data, sizeof(Keyboard_Data), 1, RAMFile);
            fwrite(&Keyboard_Enabled, sizeof(Keyboard_Enabled), 1, RAMFile);
            fwrite(&RTC, sizeof(RTC_Ty), 1, RAMFile);
            fwrite(&SCC, sizeof(SCC_Ty), 1, RAMFile);
            fwrite(&SCC_Mode, sizeof(SCC_Mode), 1, RAMFile);
            fwrite(&SCC_Reg, sizeof(SCC_Reg), 1, RAMFile);
            fwrite(&Sony_Installed, sizeof(Sony_Installed), 1, RAMFile);
            fwrite(&Sony_Opened, sizeof(Sony_Opened), 1, RAMFile);
            fwrite(Sony_Table, sizeof(Sony_Table), 1, RAMFile);
            fwrite(&ParamBlk, sizeof(ParamBlk), 1, RAMFile);
            fwrite(&DeviceCtl, sizeof(DeviceCtl), 1, RAMFile);
            fwrite(&Floppy, sizeof(Floppy), 1, RAMFile);
            fwrite(&Sound, sizeof(Sound_Ty), 1, RAMFile);
            fwrite(&Installed, sizeof(Installed), 1, RAMFile);
            fwrite(&VIA, sizeof(VIA_Ty), 1, RAMFile);
            fwrite(&T1_Active, sizeof(T1_Active), 1, RAMFile);
            fwrite(&T2_Active, sizeof(T2_Active), 1, RAMFile);
            fwrite(&SR_Active, sizeof(SR_Active), 1, RAMFile);
            fwrite(&SR_Timer_Active, sizeof(SR_Timer_Active), 1, RAMFile);
            fwrite(&SR_Time, sizeof(SR_Time), 1, RAMFile);

            // Dump Macintosh Memory [16mb]
            for (i=0; i<0x1000000; i+=4) {
               x=get_long(i);
               fwrite(&x, 4, 1, RAMFile);
            }

            // For some odd reason, the cpu state gets screwed in the mem dump
            fseek(RAMFile, 0L, SEEK_SET);
            fread(&regs, sizeof(regs), 1, RAMFile);
            fread(&lastint_regs, sizeof(lastint_regs), 1, RAMFile);

            fclose(RAMFile);
                    
            break;
         }  

         case '2': {
            FILE *RAMFile;
            char *MacRAM = (char *) get_real_address (kRAM_Size-kRAM_Size);
            CPTR i;
            CPTR z;
            ULONG x;

            RAMFile = fopen("vMac.RAM", "rb+");

            // Load CPU state
            fread(&regs, sizeof(regs), 1, RAMFile);
            fread(&lastint_regs, sizeof(lastint_regs), 1, RAMFile);

            // Load Hardware chips information and misc. control variables
            fread(&IWM, sizeof(IWM_Ty), 1, RAMFile);
            fread(&Keyboard_Data, sizeof(Keyboard_Data), 1, RAMFile);
            fread(&Keyboard_Enabled, sizeof(Keyboard_Enabled), 1, RAMFile);
            fread(&RTC, sizeof(RTC_Ty), 1, RAMFile);            
            fread(&SCC, sizeof(SCC_Ty), 1, RAMFile);
            fread(&SCC_Mode, sizeof(SCC_Mode), 1, RAMFile);
            fread(&SCC_Reg, sizeof(SCC_Reg), 1, RAMFile);
            fread(&Sony_Installed, sizeof(Sony_Installed), 1, RAMFile);
            fread(&Sony_Opened, sizeof(Sony_Opened), 1, RAMFile);
            fread(Sony_Table, sizeof(Sony_Table), 1, RAMFile);
            fread(&ParamBlk, sizeof(ParamBlk), 1, RAMFile);
            fread(&DeviceCtl, sizeof(DeviceCtl), 1, RAMFile);
            fread(&Floppy, sizeof(Floppy), 1, RAMFile);
            fread(&Sound, sizeof(Sound_Ty), 1, RAMFile);
            fread(&Installed, sizeof(Installed), 1, RAMFile);
            fread(&VIA, sizeof(VIA_Ty), 1, RAMFile);
            fread(&T1_Active, sizeof(T1_Active), 1, RAMFile);
            fread(&T2_Active, sizeof(T2_Active), 1, RAMFile);
            fread(&SR_Active, sizeof(SR_Active), 1, RAMFile);
            fread(&SR_Timer_Active, sizeof(SR_Timer_Active), 1, RAMFile);
            fread(&SR_Time, sizeof(SR_Time), 1, RAMFile);

            // Load Macintosh Memory [16mb]
            ROMmem_write_enable_set(on);
            for (i=0; i<0x1000000; i+=4) {
               fread(&x, 4, 1, RAMFile);
               put_long(i, x);
            }
            ROMmem_write_enable_set(off);

            // For some odd reason, the cpu state gets screwed in the mem load
            fseek(RAMFile, 0L, SEEK_SET);
            fread(&regs, sizeof(regs), 1, RAMFile);
            fread(&lastint_regs, sizeof(lastint_regs), 1, RAMFile);

            fclose(RAMFile);
                    
            break;
         }  

         case 'V': {
            VGA_ViewScreen();
            break;
         }

         case 'M': {
            MacInterrupt();
            break;
         }

         case 'R': {
            extern void Reset_Emulator(void);

            Reset_Emulator();

            break;
         }

         case 'I': {
            char z;
            int i=0;
            char x[12];

            if (more_params(&inptr)) {
               z=next_char(&inptr);
               while (z!='\0') {
                  x[i++]=z;
                  z=next_char(&inptr);
               }   
               x[i]='\0';

               if (inStream!=stdin)
                  fclose(inStream);

               if ((inStream=fopen(x, "rt"))==NULL) {
                  fprintf(stderr, "can't find %s.  Redirection aborted.\n", x);
                  inStream=stdin;
               } else
                  fprintf(stderr, "stdin redirected to %s.\n", x);
            } else {
               if (inStream!=stdin)
                  fclose(inStream);

               inStream=stdin;
               fprintf(stderr, "stdin redirection disabled.\n");
            }

            break;
         }

         case 'O': {
            char z;
            int i=0;
            char x[12];

            if (more_params(&inptr)) {
               z=next_char(&inptr);
               while (z!='\0') {
                  x[i++]=z;
                  z=next_char(&inptr);
               }   
               x[i]='\0';

               if (ouStream!=stdout)
                  fclose(ouStream);

               ouStream=fopen(x, "wt+");
               fprintf(stderr, "stdout redirected to %s.\n", x);
            } else {
               if (ouStream!=stdout)
                  fclose(ouStream);

               ouStream=stdout;
               fprintf(stderr, "stdout redirection disabled.\n");
            }

            break;
         }

         case 'S': {
            char *args[] = {
               "command", 0
            };

            _set_screen_lines(25);
            clrscr();

            fprintf(ouStream, "vMac: Spawning a DOS command shell.. \n");
            spawnvp(P_WAIT, "command", args);

            _set_screen_lines(50);
            clrscr();
            fprintf(ouStream, "debugging...\n");
            m68k_dumpstate(&nextpc);
            nxdis = nextpc; nxmem = 0;

            break;
         }

         case 'Q': {
            if (ouStream!=stdout) {
               fclose(ouStream);
               ouStream=stdout;
            }

            if (inStream!=stdin) {
               fclose(inStream);
               inStream=stdin;
            }

            return;
         }

         case 'h':
         case '?': {
            fprintf (ouStream, "          HELP for vMac Debugger                                              \n");
            fprintf (ouStream, "         -----------------------                                            \n\n");
            fprintf (ouStream, "  h,?:                  Show this help page.                                  \n");

            fprintf (ouStream, "  m <address> <lines>:  Memory dump starting at <address>.                    \n");
            fprintf (ouStream, "  d <address> <lines>:  Disassembly starting at <address>.                    \n");
            fprintf (ouStream, "  w <address> <value>:  Write a value (byte size) at <address>.               \n");
            fprintf (ouStream, "  W <reg> <value>:      Write a value (long size) to <reg>.                   \n");
            fprintf (ouStream, "  i:                    Enable/Disable ROM write access.                      \n");
            
            fprintf (ouStream, "  r:                    Dump state of the CPU.                                \n");
            fprintf (ouStream, "  g <address>:          Start execution at the current address or <address>.  \n");
            fprintf (ouStream, "  t:                    Step one instruction.                                 \n");
            fprintf (ouStream, "  z:                    Step through one instruction.                         \n");
            fprintf (ouStream, "  Z <count>:            Step forward <count> instructions.                    \n");
            fprintf (ouStream, "  f <address>:          Step forward until PC == <address>.                   \n");
            fprintf (ouStream, "  b <address>:          Set a break-point at the given <address>.             \n");

            fprintf (ouStream, "  1:                    Dump the RAM and CPU state into ~vMac.RAM~.           \n");
            fprintf (ouStream, "  2:                    Load the RAM and CPU state from ~vMac.RAM~.           \n");
            fprintf (ouStream, "  V:                    View screen contents.                                 \n");

            fprintf (ouStream, "  M:                    Interrupt the Macintosh Plus.                         \n");
            fprintf (ouStream, "  R:                    Restart the Macintosh Plus.                           \n");

            fprintf (ouStream, "  I <filename>:         Redirect stdin to <filename>.                         \n");
            fprintf (ouStream, "  O <filename>:         Redirect stdout to <filename>.                        \n");
            fprintf (ouStream, "  S:                    Shell to DOS.                                         \n");
            fprintf (ouStream, "  Q:                    Quit the emulator.                                  \n\n");
            break;
         }
      }
   }
}
